class OrderVerificationController < ApplicationController

    FRAUD_FILTERS = ['NOT USED',
            'AVS No Match',
            'AVS Partial Match',
            'AVS Unavailable/Unsupported',
            'Card Security Code (CSC) Mismatch',
            'Maximum Transaction Amount',
            'Unconfirmed Address',
            'Country Monitor',
            'Large Order Number',
            'Billing/Shipping Address Mismatch',
            'Risky Zip Code',
            'Suspected Freight Forwarder Check',
            'Total Purchase Price Minimum',
            'IP Address Velocity',
            'Risky Email Address Domain Check',
            'Risky Bank Identification Number (BIN) Check',
            'Risky IP Address Range',
            'Paypayl Fraud Model']

    protected

    # If the order is valid, the valid order object will
    # be returned with the txn_id saved down. Otherwise,
    # null will be returned.
    def verify_order(params)

        logger.info "Begin verify order"

        # Note mismatches. Each node of the array will contain
        # the parameter name, the passed value, and the expected
        # value, in that order.
        # mismatches << [ "param_name", my_expected_value, params["param_name"] ]
        mismatches = []

        order = Order.find_by_id(params["invoice"])

        # Verify the transaction level, business related details
        return if !verify_txn_business_details(params, mismatches)

        # Verify the order level details
        return if !verify_txn_order_details(order, params, mismatches)

        # Verify the customer level details
        return if !verify_txn_customer_details(order, params, mismatches)

        # Verify the order item details
        return if !verify_txn_order_item_details(order, params, mismatches)

        # Make sure we don't have any mismatches before we save this data down.
        if mismatches.size > 0
            note_mismatches order, mismatches
            return
        end

        # Verify and take action on the payment status
        return if !verify_txn_payment_status(order, params)

        # Verify transactional details and save any changed data
        return if !verify_and_save_txn_details(order, params, mismatches)

        logger.info "End verify order"

        return Order.find_by_id(params["invoice"])
    end

    def verify_txn_business_details(params, mismatches)
        # Make sure the business credential are ours
        check_equality(Util.paypal_business_account, params, "receiver_id", mismatches)
        check_equality(Util.paypal_business_email, params, "business", mismatches)
        check_equality(Util.paypal_business_email, params, "receiver_email", mismatches)

        logger.info "Done checking business credentials"

        return true
    end

    def verify_txn_payment_status(order, params)
        if params["payment_status"].nil? || params["payment_status"] == ""
            logger.error "The payment status was not found."
            note_mismatches order, mismatches
            return false
        end

        # We want to take special actions on certain payment statuses
        case params["payment_status"]
            when "Completed", "Canceled_Reversal"
                # In this scenario we need to flag an order as available
                # to be shipped again
                order.pending = false

            when "Denied", "Expired", "Failed", "Refunded", "Reversed", "Voided", "Processed", "Pending"
                # Note that it is NOT ok to ship this order yet
                order.pending = true

            else
                logger.error "The payment status is unknown!"
                mismatches << [ "payment_status", "Something valid...", params["payment_status"]]
                note_mismatches order, mismatches
                return false
        end


        case params["payment_status"]
            when "pending_reason"
                order.pending_reason = params["pending_reason"]
            when "Reversed", "Refunded", "Cancelled_Reversal"
                order.pending_reason = params["reason_code"]
            else
                order.pending_reason = ""
        end

        if !params["case_id"].nil?
            order.pending_reason = params["reason_code"]
        end

        n = 1
        fraud_reasons = ""
        while params["Fraud_Management_Filters_" + n.to_s]
            fraud_reasons += n.to_s + ": " + FRAUD_FILTERS[params["Fraud_Management_Filters_" + n.to_s]] + "\n"
        end

        if fraud_reasons.size > 0
            order.risky = true
            order.risky_details = fraud_reasons
        else
            order.risky = false
            order.risky_details = ""
        end

        order.save

        logger.info "Done checking payment status"
    end

    def verify_txn_order_details(order, params, mismatches)

        if order.nil?
            logger.error "The order was not found. Requested order: " + params["invoice"]
            note_mismatches order, mismatches
            return false
        end

        logger.info "Found order"

        # Make sure the order level details are correct
        check_equality(order.total, params, "mc_gross", mismatches)
        check_equality(order.shipping, params, "mc_shipping", mismatches)
        check_equality(order.tax, params, "tax", mismatches)
        check_equality(order.handling, params, "handling", mismatches)

        # IPN doesn't have this attribute so we can ignore it if it isn't present
        if !params["num_cart_items"].nil?
            check_equality(order.products.size.to_s, params, "num_cart_items", mismatches)
        end

        logger.info "Verified order details"

        return true
    end

    def verify_txn_order_item_details(order, params, mismatches)

        # Loop through each item and verify the details
        i = 1

        until params["item_name" + i.to_s].nil?

            product = Product.find_by_id(params["item_number" + i.to_s])
            if product.nil?
                logger.error "Product was not found. Expected product: " + params["item_number" + i.to_s]
                note_mismatches order, mismatches
                return false
            end

            # Verify the details match up
            check_equality(product.name, params, "item_name" + i.to_s, mismatches)
            check_equality(product.shipping, params, "mc_shipping" + i.to_s, mismatches)
            check_equality(product.price + product.shipping, params, "mc_gross_" + i.to_s, mismatches)
            check_equality(product.handling, params, "mc_handling" + i.to_s, mismatches)
            check_equality(product.quantity, params, "quantity" + i.to_s, mismatches)

             # IPN doesn't have this attribute so we can ignore it if it isn't present
            if !params["mc_tax" + i.to_s].nil?
                check_equality(product.tax, params, "mc_tax" + i.to_s, mismatches)
            end

            logger.info "Verified product " + i.to_s

            i += 1
        end

        if order.products.size != (i - 1)
            logger.error "The number of products on the order does not match."
            logger.error "Expected count: " + order.products.size.to_s
            logger.error "Actual count: " + (i - 1)
            note_mismatches order, mismatches
            return false
        end

        logger.info "Verified order item details"

        return true
    end

    def verify_txn_customer_details(order, params, mismatches)

        # Make sure the customer matches the customer on
        # the order.
        customer = Customer.find_by_id(params["custom"])
        if customer.nil?
            note_verification_failure "Customer is not found. Customer requested: " + params["custom"], request
            note_mismatches order, mismatches
            return false
        end

        if !Util.values_equal(order.customer, customer)
            mismatches << [ "customer", customer.id, order.customer.id ]
        end

        logger.info "Found customer"

        # Verify the customer information
        check_equality(customer.first, params, "first_name", mismatches)
        check_equality(customer.last, params, "last_name", mismatches)
        check_equality(customer.email, params, "payer_email", mismatches)

        # IPN doesn't have this attribute so we can ignore it if it isn't present
        if !params["address_street"].nil?
            check_equality(customer.address, params, "address_street", mismatches)
        end

        # IPN doesn't have this attribute so we can ignore it if it isn't present
        if !params["address_city"].nil?
            check_equality(customer.city, params, "address_city", mismatches)
        end

        # IPN doesn't have this attribute so we can ignore it if it isn't present
        if !params["address_state"].nil?
            check_equality(customer.state, params, "address_state", mismatches)
        end

        # IPN doesn't have this attribute so we can ignore it if it isn't present
        if !params["address_zip"].nil?
            check_equality(customer.zip, params, "address_zip", mismatches)
        end

        # IPN doesn't have this attribute so we can ignore it if it isn't present
        if !params["address_country"].nil?
            check_equality(customer.country, params, "address_country", mismatches)
        end

        logger.info "Verified customer information"

        return true
    end

    def verify_and_save_txn_details(order, params, mismatches)
        # Update whether Paypal was able to confirm the buyer
        # address or not
        # IPN doesn't have this attribute so we can ignore it if it isn't present

        if !params["address_status"].nil?
            customer = Customer.find_by_id(params["custom"])
            customer.address_confirmed = params["address_status"]
            customer.save
            logger.info "Saved customer details"
        end

        # Update order details
        if order.pp_txn_fee.nil?
            order.pp_txn_fee = params["mc_fee"].to_f
        else
            check_equality(order.pp_txn_fee, params, "mc_fee", mismatches)
        end

        if order.pp_txn_id.nil?
            order.pp_txn_id = params["txn_id"]
        else
            check_equality(order.pp_txn_id, params, "txn_id", mismatches)
        end

        # This isn't sent through IPA
        if !params["receipt_id"].nil?
            if order.pp_receipt_id.nil?
                order.pp_receipt_id = params["receipt_id"]
            else
                check_equality(order.pp_receipt_id, params, "receipt_id", mismatches)
            end
        end

        if order.payment_status.nil?
            order.payment_status = params["payment_status"]
        else
            check_equality(order.payment_status, params, "payment_status", mismatches)
        end

        if !params["protection_eligibility"].nil?
            order.pp_protection_eligible = params["protection_eligibility"]
        end

        if !params["payment_type"].nil?
            order.payment_type = params["payment_type"]
        end

        if !params["auth"].nil?
            order.pp_auth = params["auth"]
        end

        if !params["auth_amount"].nil?
            order.authorization_amount = params["auth_amount"]
        end

        if !params["auth_exp"].nil?
            order.authorization_expires = params["auth_exp"]
        end

        if !params["auth_id"].nil?
            order.authorization_id = params["auth_id"]
        end

        if !params["auth_status"].nil?
            order.authorization_status = params["auth_status"]
        end

        if !params["exchange_rate"].nil?
            order.exchange_rate = params["exchange_rate"]
        end

        if !params["payer_status"].nil?
            order.payer_status = params["payer_status"]
        end

        if !params["remaining_settle"].nil?
            order.remaining_settle = params["remaining_settle"]
        end

        if !params["settle_amount"].nil?
            order.settle_amount = params["settle_amount"]
        end

        if !params["settle_currency"].nil?
            order.settle_currency = params["settle_currency"]
        end

        if !params["transaction_entity"].nil?
            order.transaction_entity = params["transaction_entity"]
        end

        if !params["case_id"].nil?
            order.dispute_case_id = params["case_id"]
        end

        if !params["case_type"].nil?
            order.dispute_case_type = params["case_type"]
        end

        if !params["case_creation_date"].nil?
            order.dispute_case_creation_date = params["case_creation_date"]
        end

        if mismatches.length > 0
            note_mismatches(order, mismatches)
            return false
        end

        if !order.risky
            order.risky_details = ""
        end

        order.save
        logger.info "Saved order details"

        return true
    end

    def check_equality(val, params, param_arg, mismatches)
        if !Util.values_equal(val, params[param_arg])
            mismatches << [ param_arg, val, params[param_arg]]
        end
    end

    def note_mismatches(order, mismatches)

        logger.error "The following mismatches were found:"

        to_log = ""
        for vals in mismatches
            to_log += vals[0].to_s + ", Expected: " + vals[1].to_s + ", Actual: " + vals[2].to_s + "\n"
        end

        logger.error to_log

        # If order is not nil, note that the values were not right on the
        # order itself so that the turner will be cautious
        if !order.nil?
            order.risky = true

            if to_log.size > 4000
                truncate(to_log, 4000)
            end

            order.risky_details = to_log
            order.save
        end

        logger.error "No changes were made to the order."
    end

    def notify_admin(params)
        ApplicationMailer.deliver_parameter_email("IPN - TXN NOTIFICATION (" + params["txn_type"] + ")", params)
        logger.error "Notifying admin of transaction type " + params["txn_type"]
    end

    def unexpected_payment_source(params)
        ApplicationMailer.deliver_parameter_email("IPN - TXN UNEXPECTED PAYMENT SOURCE (" + params["txn_type"] + ")", params)
        logger.error "Notifying admin of an unexpected payment source: " + params["txn_type"]
    end
end